/*
 *
 *	Utility for converting quota file from old to new format
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <linux/quota.h>

#include <asm/byteorder.h>

#include "common.h"
#include "quota_io.h"
#include "quotasys.h"

#define OLDNAME "quota"

struct old_dqblk {
	__u32 dqb_bhardlimit;	/* absolute limit on disk blks alloc */
	__u32 dqb_bsoftlimit;	/* preferred limit on disk blks */
	__u32 dqb_curblocks;	/* current block count */
	__u32 dqb_ihardlimit;	/* absolute limit on allocated inodes */
	__u32 dqb_isoftlimit;	/* preferred inode limit */
	__u32 dqb_curinodes;	/* current # allocated inodes */
	time_t dqb_btime;	/* time limit for excessive disk use */
	time_t dqb_itime;	/* time limit for excessive inode use */
};

int type;
char *mntpoint;

void parse_options(int argcnt, char **argstr)
{
	int ret;
	char *slash = strrchr(argstr[0], '/'), cmdname[MAXFILENAMELEN];

	if (!slash)
		slash = argstr[0];
	else
		slash++;
	sstrncpy(cmdname, slash, MAXFILENAMELEN);
	while ((ret = getopt(argcnt, argstr, "Vugh:")) != EOF) {
		switch (ret) {
			case '?':
			case 'h':
usage:
				printf("Utility for converting quota files.\nUsage:\n%s [-u] [-g] mountpoint\n", cmdname);
				printf("Bugs to %s\n", MY_EMAIL);
				exit(1);
			case 'V':
				version();
				exit(0);
			case 'u':
				type |= QS_USRQUOTA;
				break;
			case 'g':
				type |= QS_GRPQUOTA;
				break;
		}
	}
	if (optind + 1 != argcnt) {
		puts("Bad number of arguments.");
		goto usage;
	}
	if (!(type & (QS_USRQUOTA | QS_GRPQUOTA)))
		type |= QS_USRQUOTA;
	mntpoint = argstr[optind];
}

void create_name(char *buf, int type)
{
	sstrncpy(buf, mntpoint, MAXFILENAMELEN);
	if (buf[strlen(buf) - 1] != '/')
		sstrncat(buf, "/", MAXFILENAMELEN);
	sstrncat(buf, OLDNAME, MAXFILENAMELEN);
	sstrncat(buf, ".", MAXFILENAMELEN);
	sstrncat(buf, type2name(type), MAXFILENAMELEN);
}

/* Do the conversion */
void convert_entries(int ofd)
{
	struct old_dqblk oblk;
	struct dquot dquot;
	int ret;
	qid_t id = 0;

	lseek(ofd, 0, SEEK_SET);
	while ((ret = read(ofd, &oblk, sizeof(struct old_dqblk)))) {
		if (ret < 0)
			die(2, "Error while reading old quota file: %s\n", strerror(errno));
		dquot.dq_id = id;
		dquot.dq_off = 0;
		dquot.dq_dqb.dqb_bsoftlimit = oblk.dqb_bsoftlimit;
		dquot.dq_dqb.dqb_bhardlimit = oblk.dqb_bhardlimit;
		dquot.dq_dqb.dqb_isoftlimit = oblk.dqb_isoftlimit;
		dquot.dq_dqb.dqb_ihardlimit = oblk.dqb_ihardlimit;
		dquot.dq_dqb.dqb_curspace = oblk.dqb_curblocks * QUOTABLOCK_SIZE;
		dquot.dq_dqb.dqb_curinodes = oblk.dqb_curinodes;
		if (id) {	/* In structure of UID 0 were stored graces.. :-| */
			dquot.dq_dqb.dqb_btime = oblk.dqb_btime;
			dquot.dq_dqb.dqb_itime = oblk.dqb_itime;
		}
		else {
			info.dqi_bgrace = oblk.dqb_btime;
			info.dqi_igrace = oblk.dqb_itime;
			dquot.dq_dqb.dqb_itime = 0;
			dquot.dq_dqb.dqb_btime = 0;
			update_times(&dquot.dq_dqb);
		}
		commit_dquot(&dquot, 1);
		id++;
	}
}

void convert_file(int type)
{
	char name_buf[MAXFILENAMELEN], new_name_buf[MAXFILENAMELEN];
	int fd;

	create_name(name_buf, type);
	if ((fd = open(name_buf, O_RDONLY)) < 0)
		die(1, "Can't open source file %s: %s\n", name_buf, strerror(errno));
	new_io(mntpoint, "/dev/null", type);	/* Device won't be needed */
	convert_entries(fd);
	free_io();
	close(fd);
	get_qfname(mntpoint, type, name_buf);
	strcpy(new_name_buf, name_buf);
	sstrncat(new_name_buf, ".new", MAXFILENAMELEN);
	if (rename(new_name_buf, name_buf) < 0)
		die(2, "Can't rename new quotafile %s to name %s: %s\n", new_name_buf, name_buf, strerror(errno));
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);
	if (type & QS_USRQUOTA)
		convert_file(QS_USRQUOTA);
	if (type & QS_GRPQUOTA)
		convert_file(QS_GRPQUOTA);
	return 0;
}

