/*
 *
 *	IO operations in quota files
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/file.h>

#include <linux/quota.h>

#include <asm/byteorder.h>
#include <asm/types.h>

#include "quota_io.h"
#include "quotasys.h"
#include "common.h"

struct mem_dqinfo info;	/* Info from file */
int io_flags;		/* Flags */
static int file_fd;		/* Handle of manipulated file */
static short quota_type;	/* Type of quota */
static char quota_dev[MAXDEVLEN];	/* Device we operate on */
static uint magics[MAXQUOTAS] = INITQMAGICS;	/* Magics of quota files */
static int known_versions[MAXQUOTAS] = INITKNOWNVERSIONS;	/* Latest known versions */

dqbuf_t getdqbuf(void)
{
	dqbuf_t buf = malloc(DQBLKSIZE);
	if (!buf) {
		fputs("Not enough memory for io buffer.\n", stderr);
		exit(3);
	}
	return buf;
}

inline void freedqbuf(dqbuf_t buf)
{
	free(buf);
}

int empty_dquot(struct disk_dqblk *dq)
{
	static struct disk_dqblk fake_dquot;

	return !memcmp(dq, &fake_dquot, sizeof(struct disk_dqblk));
}

/* Read info block from file */
static int read_info(void)
{
	struct disk_dqinfo dinfo;
	int len;

	lseek(file_fd, DQINFOOFF, SEEK_SET);
	len = read(file_fd, &dinfo, sizeof(struct disk_dqinfo));
	if (len < 0)
		return -errno;
	if (len != sizeof(struct disk_dqinfo))
		return -ENODATA;
	info.dqi_bgrace = __le32_to_cpu(dinfo.dqi_bgrace);
	info.dqi_igrace = __le32_to_cpu(dinfo.dqi_igrace);
	info.dqi_flags = __le32_to_cpu(dinfo.dqi_flags);
	info.dqi_blocks = __le32_to_cpu(dinfo.dqi_blocks);
	info.dqi_free_blk = __le32_to_cpu(dinfo.dqi_free_blk);
	info.dqi_free_entry = __le32_to_cpu(dinfo.dqi_free_entry);
	return 0;
}

/*
 *	Initialize all data for further io operations - MUST be called before any other io operation
 */
void init_io(char *mntpoint, char *dev, int type)
{
	int err;
	char name_buf[MAXFILENAMELEN];
	struct disk_dqheader head;

	quota_type = user2kernquota(type);
	sstrncpy(quota_dev, dev, MAXDEVLEN);
	if (quotactl(QCMD(Q_GETINFO, quota_type), dev, 0, (void *)&info) < 0) { /* Is quota enabled? */
		if (errno != ENOSYS && errno != ESRCH && errno != EINVAL)	/* ! Quota not in kernel or not turned on? */
			die(4, "Error while getting base info from quotactl(): %s\nMaybe old kernel?\n", strerror(errno));
	}
	else
		io_flags |= IOFL_QUOTAON;

	get_qfname(mntpoint, type, name_buf);
	if ((file_fd = open(name_buf, QIO_ENABLED ? O_RDONLY : O_RDWR)) < 0)
		die(2, "Can't open quota file %s: %s\n", name_buf, strerror(errno));
	if (!QIO_ENABLED)
		flock(file_fd, LOCK_EX);	/* Lock file to prevent races with other utilities */
	err = read(file_fd, &head, sizeof(struct disk_dqheader));
	if (err != sizeof(struct disk_dqheader)) {
		if (err >= 0)
			errno = ENODATA;
		die(2, "Can't read quotafile header: %s\n", strerror(errno));
	}
	if (__le32_to_cpu(head.dqh_magic) != magics[quota_type])
		die(5, "File %s is not quotafile.\n", name_buf);
	if (__le32_to_cpu(head.dqh_version) > known_versions[quota_type])
		die(5, "Sorry but I can't handle file with version %d (max %d).\n",
		  __le32_to_cpu(head.dqh_version), known_versions[quota_type]);
	if (!QIO_ENABLED)
		if ((err = read_info()) < 0)
			die(2, "Can't read base info from file: %s\n", strerror(-err));
}

/* Writes identification header */
static void write_qf_header(int fd, int type)
{
	int magics[MAXQUOTAS] = INITQMAGICS;
	int versions[MAXQUOTAS] = INITKNOWNVERSIONS;
	struct disk_dqheader head = {__cpu_to_le32(magics[user2kernquota(type)]), __cpu_to_le32(versions[user2kernquota(type)])};
	int ret;

	lseek(fd, 0, SEEK_SET);
	ret = write(fd, &head, sizeof(struct disk_dqheader));
	if (ret != sizeof(struct disk_dqheader)) {
		if (ret >= 0)
			errno = ENOSPC;
		die(2, "Can't write quota header: %s\n", strerror(errno));
	}
}


/*
 *	Initialize io for new file
 */
void new_io(char *mntpoint, char *dev, int type)
{
	char name_buf[MAXFILENAMELEN];

	get_qfname(mntpoint, type, name_buf);
	sstrncat(name_buf, ".new", MAXFILENAMELEN);
	if ((file_fd = open(name_buf, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR)) < 0)
		die(2, "Can't create destination file %s: %s\n", name_buf, strerror(errno));
	flock(file_fd, LOCK_EX);
	quota_type = user2kernquota(type);
	sstrncpy(quota_dev, dev, MAXDEVLEN);
	write_qf_header(file_fd, type);
	info.dqi_flags = 0;
	info.dqi_blocks = DQTREEOFF + 1;
	info.dqi_free_blk = 0;
	info.dqi_free_entry = 0;
	info.dqi_bgrace = MAX_DQ_TIME;
	info.dqi_igrace = MAX_IQ_TIME;
	mark_quotafile_info_dirty();
	lseek(file_fd, info.dqi_blocks * DQBLKSIZE - 1, SEEK_SET);
	name_buf[0] = 0;
	write(file_fd, name_buf, 1);
}

/*
 *	Write info to file
 */
void commit_info(void)
{
	struct disk_dqinfo dinfo;
	int err;

	if (QIO_ENABLED) {
		if (quotactl(QCMD(Q_SETINFO, quota_type), quota_dev, 0, (void *)&info) < 0)
			die(4, "Can't set kernel information by quotactl(): %s\n", strerror(errno));
		return;
	}
	dinfo.dqi_bgrace = __cpu_to_le32(info.dqi_bgrace);
	dinfo.dqi_igrace = __cpu_to_le32(info.dqi_igrace);
	dinfo.dqi_flags = __cpu_to_le32(info.dqi_flags & DQF_MASK);
	dinfo.dqi_blocks = __cpu_to_le32(info.dqi_blocks);
	dinfo.dqi_free_blk = __cpu_to_le32(info.dqi_free_blk);
	dinfo.dqi_free_entry = __cpu_to_le32(info.dqi_free_entry);
	lseek(file_fd, DQINFOOFF, SEEK_SET);
	err = write(file_fd, &dinfo, sizeof(struct disk_dqinfo));
	if (err != sizeof(struct disk_dqinfo)) {
		if (err >= 0)
			errno = ENOSPC;
		die(2, "Can't write base quota information: %s\n", strerror(errno));
	}
}

/*
 *	Free all io structures and eventually write them
 */
void free_io(void)
{
	if (io_flags & IOFL_INFODIRTY)
		commit_info();
	if (!QIO_ENABLED)
		flock(file_fd, LOCK_UN);
	close(file_fd);
	io_flags = 0;
}

void disk2memdqb(struct mem_dqblk *M, struct disk_dqblk *D)
{
	M->dqb_bhardlimit = __le32_to_cpu(D->dqb_bhardlimit);
	M->dqb_bsoftlimit = __le32_to_cpu(D->dqb_bsoftlimit);
	M->dqb_curspace = __le64_to_cpu(D->dqb_curspace);
	M->dqb_ihardlimit = __le32_to_cpu(D->dqb_ihardlimit);
	M->dqb_isoftlimit = __le32_to_cpu(D->dqb_isoftlimit);
	M->dqb_curinodes = __le32_to_cpu(D->dqb_curinodes);
	/* We might do this somehow better for archs where sizeof(time_t) == 4 */
	M->dqb_btime = __le64_to_cpu(D->dqb_btime);
	M->dqb_itime = __le64_to_cpu(D->dqb_itime);
}

void mem2diskdqb(struct disk_dqblk *D, struct mem_dqblk *M, qid_t id)
{
	D->dqb_id = __cpu_to_le32(id);
	D->dqb_bhardlimit = __cpu_to_le32(M->dqb_bhardlimit);
	D->dqb_bsoftlimit = __cpu_to_le32(M->dqb_bsoftlimit);
	D->dqb_curspace = __cpu_to_le64(M->dqb_curspace);
	D->dqb_ihardlimit = __cpu_to_le32(M->dqb_ihardlimit);
	D->dqb_isoftlimit = __cpu_to_le32(M->dqb_isoftlimit);
	D->dqb_curinodes = __cpu_to_le32(M->dqb_curinodes);
	/* We might do this somehow better for archs where sizeof(time_t) == 4 */
	D->dqb_btime = __cpu_to_le64(M->dqb_btime);
	D->dqb_itime = __cpu_to_le64(M->dqb_itime);
}


static void get_sys_quota(qid_t id, struct dquot *dquot)
{
	if (quotactl(QCMD(Q_GETQUOTA, quota_type), quota_dev, id, (void *)&dquot->dq_dqb) < 0)
		die(4, "Error while getting structure for %d by quotactl(): %s\n", strerror(errno));
}

/* Read given block */
void read_blk(uint blk, dqbuf_t buf)
{
	int err;

	lseek(file_fd, blk << DQBLKSIZE_BITS, SEEK_SET);
	err = read(file_fd, buf, DQBLKSIZE);
	if (err < 0)
		die(2, "Can't read block %u: %s\n", blk, strerror(errno));
	else if (err != DQBLKSIZE)
		memset(buf + err, 0, DQBLKSIZE - err);
}

#define GETENTRIES(buf) ((struct disk_dqblk *)(((char *)(buf)) + sizeof(struct disk_dqdbheader)))

/* 
 *	Write block
 */
int write_blk(uint blk, dqbuf_t buf)
{
	int err;

	lseek(file_fd, blk << DQBLKSIZE_BITS, SEEK_SET);
	err = write(file_fd, buf, DQBLKSIZE);
	if (err < 0 && errno != ENOSPC)
		die(2, "Can't write quota block (%u): %s\n", blk, strerror(errno));
	if (err != DQBLKSIZE)
		return -ENOSPC;
	return 0;
}

static int get_free_dqblk(void)
{
	dqbuf_t buf = getdqbuf();
	struct disk_dqdbheader *dh = (struct disk_dqdbheader *)buf;
	int blk;

	if (info.dqi_free_blk) {
		blk = info.dqi_free_blk;
		read_blk(blk, buf);
		info.dqi_free_blk = __le32_to_cpu(dh->dqdh_next_free);
	}
	else {
		memset(buf, 0, DQBLKSIZE);
		if (write_blk(info.dqi_blocks, buf) < 0) {	/* Assure block allocation... */
			freedqbuf(buf);
			fprintf(stderr, "Can't allocate new quota block (out of disk space).\n");
			return -ENOSPC;
		}
		blk = info.dqi_blocks++;
	}
	mark_quotafile_info_dirty();
	freedqbuf(buf);
	return blk;
}

static void put_free_dqblk(dqbuf_t buf, uint blk)
{
	struct disk_dqdbheader *dh = (struct disk_dqdbheader *)buf;

	dh->dqdh_next_free = __cpu_to_le32(info.dqi_free_blk);
	dh->dqdh_prev_free = __cpu_to_le32(0);
	dh->dqdh_entries = __cpu_to_le16(0);
	info.dqi_free_blk = blk;
	mark_quotafile_info_dirty();
	write_blk(blk, buf);
}

/* Remove given block from the list of blocks with free entries */
static void remove_free_dqentry(dqbuf_t buf, uint blk)
{
	dqbuf_t tmpbuf = getdqbuf();
	struct disk_dqdbheader *dh = (struct disk_dqdbheader *)buf;
	uint nextblk = __le32_to_cpu(dh->dqdh_next_free), prevblk = __le32_to_cpu(dh->dqdh_prev_free);

	if (nextblk) {
		read_blk(nextblk, tmpbuf);
		((struct disk_dqdbheader *)tmpbuf)->dqdh_prev_free = dh->dqdh_prev_free;
		write_blk(nextblk, tmpbuf);
	}
	if (prevblk) {
		read_blk(prevblk, tmpbuf);
		((struct disk_dqdbheader *)tmpbuf)->dqdh_next_free = dh->dqdh_next_free;
		write_blk(prevblk, tmpbuf);
	}
	else {
		info.dqi_free_entry = nextblk;
		mark_quotafile_info_dirty();
	}
	freedqbuf(tmpbuf);
	dh->dqdh_next_free = dh->dqdh_prev_free = __cpu_to_le32(0);
	write_blk(blk, buf);	/* No matter whether write succeeds block is out of list */
}

/* Insert given block to the beginning of list with free entries */
static void insert_free_dqentry(dqbuf_t buf, uint blk)
{
	dqbuf_t tmpbuf = getdqbuf();
	struct disk_dqdbheader *dh = (struct disk_dqdbheader *)buf;

	dh->dqdh_next_free = __cpu_to_le32(info.dqi_free_entry);
	dh->dqdh_prev_free = __cpu_to_le32(0);
	write_blk(blk, buf);
	if (info.dqi_free_entry) {
		read_blk(info.dqi_free_entry, tmpbuf);
		((struct disk_dqdbheader *)tmpbuf)->dqdh_prev_free = __cpu_to_le32(blk);
		write_blk(info.dqi_free_entry, tmpbuf);
	}
	freedqbuf(tmpbuf);
	info.dqi_free_entry = blk;
	mark_quotafile_info_dirty();
}

/* Find space for dquot */
static uint find_free_dqentry(struct dquot *dquot, int *err)
{
	int blk, i;
	struct disk_dqdbheader *dh;
	struct disk_dqblk *ddquot;
	struct disk_dqblk fakedquot;
	dqbuf_t buf;

	*err = 0;
	buf = getdqbuf();
	dh = (struct disk_dqdbheader *)buf;
	ddquot = GETENTRIES(buf);
	if (info.dqi_free_entry) {
		blk = info.dqi_free_entry;
		read_blk(blk, buf);
	}
	else {
		blk = get_free_dqblk();
		if (blk < 0) {
			freedqbuf(buf);
			*err = blk;
			return 0;
		}			
		memset(buf, 0, DQBLKSIZE);
		info.dqi_free_entry = blk;
		mark_quotafile_info_dirty();
	}
	if (__le16_to_cpu(dh->dqdh_entries)+1 >= DQSTRINBLK)	/* Block will be full? */
		remove_free_dqentry(buf, blk);
	dh->dqdh_entries = __cpu_to_le16(__le16_to_cpu(dh->dqdh_entries)+1);
	memset(&fakedquot, 0, sizeof(struct disk_dqblk));
	/* Find free structure in block */
	for (i = 0; i < DQSTRINBLK && memcmp(&fakedquot, ddquot+i, sizeof(struct disk_dqblk)); i++);
	if (i == DQSTRINBLK)
		die(2, "find_free_dqentry(): Data block full but it shouldn't.\n");
	write_blk(blk, buf);
	dquot->dq_off = (blk<<DQBLKSIZE_BITS)+sizeof(struct disk_dqdbheader)+i*sizeof(struct disk_dqblk);
	freedqbuf(buf);
	return blk;
}

/* Insert reference to structure into the trie */
static int do_insert_tree(struct dquot *dquot, uint *treeblk, int depth)
{
	dqbuf_t buf;
	int newson = 0, newact = 0;
	__u32 *ref;
	uint newblk;
	int ret = 0;

	buf = getdqbuf();
	if (!*treeblk) {
		ret = get_free_dqblk();
		if (ret < 0)
			goto out_buf;
		*treeblk = ret;
		memset(buf, 0, DQBLKSIZE);
		newact = 1;
	}
	else
		read_blk(*treeblk, buf);
	ref = (__u32 *)buf;
	newblk = __le32_to_cpu(ref[GETIDINDEX(dquot->dq_id, depth)]);
	if (!newblk)
		newson = 1;
	if (depth == DQTREEDEPTH-1) {
		if (newblk)
			die(2, "Inserting already present quota entry (block %u).\n", ref[GETIDINDEX(dquot->dq_id, depth)]);
		newblk = find_free_dqentry(dquot, &ret);
	}
	else
		ret = do_insert_tree(dquot, &newblk, depth+1);
	if (newson && ret >= 0) {
		ref[GETIDINDEX(dquot->dq_id, depth)] = __cpu_to_le32(newblk);
		write_blk(*treeblk, buf);
	}
	else if (newact && ret < 0)
		put_free_dqblk(buf, *treeblk);
out_buf:
	freedqbuf(buf);
	return ret;
}

/* Wrapper for inserting quota structure into tree */
static inline void dq_insert_tree(struct dquot *dquot)
{
	int tmp = DQTREEOFF;
	if (do_insert_tree(dquot, &tmp, 0) < 0)
		die(2, "Can't write quota (id %u): No space left on device\n", (uint)dquot->dq_id);
}

/*
 *	We don't have to be afraid of deadlocks as we never have quotas on quota files...
 */
static void write_dquot(struct dquot *dquot)
{
	ssize_t ret;
	struct disk_dqblk ddquot;

	if (!dquot->dq_off)
		dq_insert_tree(dquot);
	lseek(file_fd, dquot->dq_off, SEEK_SET);
	mem2diskdqb(&ddquot, &dquot->dq_dqb, dquot->dq_id);
	ret = write(file_fd, (char *)&ddquot, sizeof(struct disk_dqblk));
	if (ret != sizeof(struct disk_dqblk)) {
		if (ret > 0)
			errno = ENOSPC;
		die(2, "Quota write failed (id %u): %s\n", (uint)dquot->dq_id, strerror(errno));
	}
}

/* Free dquot entry in data block */
static void free_dqentry(struct dquot *dquot, uint blk)
{
	struct disk_dqdbheader *dh;
	dqbuf_t buf = getdqbuf();

	if (dquot->dq_off >> DQBLKSIZE_BITS != blk)
		die(2, "Quota structure has offset to other block (%u) than it should (%u).\n", blk, (uint)(dquot->dq_off >> DQBLKSIZE_BITS));
	read_blk(blk, buf);
	dh = (struct disk_dqdbheader *)buf;
	dh->dqdh_entries = __cpu_to_le16(__le16_to_cpu(dh->dqdh_entries)-1);
	if (!__le16_to_cpu(dh->dqdh_entries)) {	/* Block got free? */
		remove_free_dqentry(buf, blk);
		put_free_dqblk(buf, blk);
	}
	else {
		memset(buf+(dquot->dq_off & ((1 << DQBLKSIZE_BITS)-1)), 0, sizeof(struct disk_dqblk));
		if (__le16_to_cpu(dh->dqdh_entries) == DQSTRINBLK-1)	/* First free entry? */
			insert_free_dqentry(buf, blk);	/* This will also write data block */
		else
			write_blk(blk, buf);
	}
	dquot->dq_off = 0;
	freedqbuf(buf);
}

/* Remove reference to dquot from tree */
static void remove_tree(struct dquot *dquot, uint *blk, int depth)
{
	dqbuf_t buf = getdqbuf();
	uint newblk;
	__u32 *ref = (__u32 *)buf;
	
	read_blk(*blk, buf);
	newblk = __le32_to_cpu(ref[GETIDINDEX(dquot->dq_id, depth)]);
	if (depth == DQTREEDEPTH-1) {
		free_dqentry(dquot, newblk);
		newblk = 0;
	}
	else
		remove_tree(dquot, &newblk, depth+1);
	if (!newblk) {
		int i;
		ref[GETIDINDEX(dquot->dq_id, depth)] = __cpu_to_le32(0);
		for (i = 0; i < DQBLKSIZE && !buf[i]; i++);	/* Block got empty? */
		if (i == DQBLKSIZE) {
			put_free_dqblk(buf, *blk);
			*blk = 0;
		}
		else
			write_blk(*blk, buf);
	}
	freedqbuf(buf);
}

/* Delete dquot from tree */
static void delete_dquot(struct dquot *dquot)
{
	uint tmp = DQTREEOFF;

	if (!dquot->dq_off)	/* Even not allocated? */
		return;
	remove_tree(dquot, &tmp, 0);
}

/* Find entry in block */
static loff_t find_block_dqentry(struct dquot *dquot, uint blk)
{
	dqbuf_t buf = getdqbuf();
	int i;
	struct disk_dqblk *ddquot = GETENTRIES(buf);

	read_blk(blk, buf);
	if (dquot->dq_id)
		for (i = 0; i < DQSTRINBLK && __le32_to_cpu(ddquot[i].dqb_id) != dquot->dq_id; i++);
	else {	/* ID 0 as a bit more complicated searching... */
		struct disk_dqblk fakedquot;

		memset(&fakedquot, 0, sizeof(struct disk_dqblk));
		for (i = 0; i < DQSTRINBLK; i++)
			if (!__le32_to_cpu(ddquot[i].dqb_id) && memcmp(&fakedquot, ddquot+i, sizeof(struct disk_dqblk)))
				break;
	}
	if (i == DQSTRINBLK)
		die(2, "Quota for id %u referenced but not present.\n", dquot->dq_id);
	freedqbuf(buf);
	return (blk << DQBLKSIZE_BITS) + sizeof(struct disk_dqdbheader) + i * sizeof(struct disk_dqblk);
}

/* Find entry for given id in the tree */
static loff_t find_tree_dqentry(struct dquot *dquot, uint blk, int depth)
{
	dqbuf_t buf = getdqbuf();
	loff_t ret = 0;
	__u32 *ref = (__u32 *)buf;

	read_blk(blk, buf);
	ret = 0;
	blk = __le32_to_cpu(ref[GETIDINDEX(dquot->dq_id, depth)]);
	if (!blk)	/* No reference? */
		goto out_buf;
	if (depth < DQTREEDEPTH-1)
		ret = find_tree_dqentry(dquot, blk, depth+1);
	else
		ret = find_block_dqentry(dquot, blk);
out_buf:
	freedqbuf(buf);
	return ret;
}

/* Find entry for given id in the tree - wrapper function */
static inline loff_t find_dqentry(struct dquot *dquot)
{
	return find_tree_dqentry(dquot, DQTREEOFF, 0);
}

void read_dquot(qid_t id, struct dquot *dquot)
{
	loff_t offset;
	ssize_t ret;
	struct disk_dqblk ddquot;

	dquot->dq_id = id;
	dquot->dq_off = 0;
	memset(&dquot->dq_dqb, 0, sizeof(struct mem_dqblk));
	if (QIO_ENABLED) {
		get_sys_quota(id, dquot);
		return;
	}
	offset = find_dqentry(dquot);
	if (offset <= 0) {	/* Entry not present? */
		dquot->dq_off = 0;
		memset(&dquot->dq_dqb, 0, sizeof(struct mem_dqblk));
	}
	else {
		dquot->dq_off = offset;
		lseek(file_fd, offset, SEEK_SET);
		ret = read(file_fd, (char *)&ddquot, sizeof(struct disk_dqblk));
		if (ret != sizeof(struct disk_dqblk)) {
			if (ret > 0)
				errno = ENOSPC;
			die(2, "Can't read quota structure for id %u: %s\n", dquot->dq_id, strerror(errno));
		}
		disk2memdqb(&dquot->dq_dqb, &ddquot);
	}
}

/* Commit changes of dquot to disk - it might also mean deleting it when quota became fake one and user has no blocks... */
void commit_dquot(struct dquot *dquot, int usage)
{
	struct mem_dqblk *b = &dquot->dq_dqb;
	if (QIO_ENABLED) {
		if (quotactl(QCMD((usage ? Q_SETQUOTA : Q_SETQLIM), quota_type), quota_dev, dquot->dq_id, (void *)&dquot->dq_dqb) < 0)
			die(4, "Can't set quota by quotactl: %s\n", strerror(errno));
		return;
	}

	if (!b->dqb_curspace && !b->dqb_curinodes && !b->dqb_bsoftlimit && !b->dqb_isoftlimit && !b->dqb_bhardlimit && !b->dqb_ihardlimit)
		delete_dquot(dquot);
	else
		write_dquot(dquot);
}

/*
 *	Sets proper dqb_itime and dqb_btime if needed
 */
void update_times(struct mem_dqblk *blk)
{
	time_t now;

	time(&now);
	if (!blk->dqb_itime)
		if (blk->dqb_curinodes > blk->dqb_isoftlimit && blk->dqb_isoftlimit)
			blk->dqb_itime = now + (time_t)info.dqi_igrace;
	if (!blk->dqb_btime)
		if ((blk->dqb_curspace + QUOTABLOCK_SIZE - 1) >> QUOTABLOCK_BITS > blk->dqb_bsoftlimit && blk->dqb_bsoftlimit)
			blk->dqb_btime = now + (time_t)info.dqi_bgrace;
}

